from typing import Generic, TypeVar

from rainbowadn.chain.stages import *
from rainbowadn.collection.linear import *
from rainbowadn.collection.pair import *
from rainbowadn.core import *
from rainbowadn.flow.core import *
from rainbowadn.flow.sequence import *
from rainbowadn.flow.verification.core import *
from rainbowadn.flow.verification.stages import *
from rainbowadn.nullability import *
from ._stackbridge import *

__all__ = ('StageBridgeVP', 'StageBridgeM', 'stage_bridge',)

BaseT = TypeVar('BaseT')
StageT = TypeVar('StageT')
HeaderT = TypeVar('HeaderT')


class StageBridgeVP(
    StageVerificationProtocol[HashPoint[BaseT], HashPoint[StageT], HashPoint[HeaderT]],
    Generic[BaseT, StageT, HeaderT]
):
    def __init__(self, stage_protocol: StageProtocol[HeaderT, BaseT, StageT], base_factory: RainbowFactory[BaseT]):
        assert isinstance(stage_protocol, StageProtocol)
        assert isinstance(base_factory, RainbowFactory)
        self.stage_protocol = stage_protocol
        self.base_factory = base_factory

    async def initial(
            self, previous: Nullable[HashPoint[BaseT]], header: HashPoint[HeaderT], stage: HashPoint[StageT]
    ) -> bool:
        assert_true(
            await self.stage_protocol.verify_header(NullableReference(previous, self.base_factory), header, stage)
        )
        return True

    async def final(self, stage: HashPoint[StageT], base: HashPoint[BaseT]) -> bool:
        assert_true(
            await self.stage_protocol.verify_state(stage, base)
        )
        return True

    async def intermediate(self, previous: HashPoint[StageT], stage: HashPoint[StageT]) -> bool:
        assert_true(
            await self.stage_protocol.verify_stage(previous, stage)
        )
        return True


class StageBridgeM(
    Mapper[
        HashPoint[Pair[BaseT, Stack[StageT]]],
        tuple[HashPoint[BaseT], Reducer[SequenceDispatcher[HashPoint[StageT], bool], bool]]
    ],
    Generic[BaseT, StageT],
):
    async def map(
            self,
            element: HashPoint[Pair[BaseT, Stack[StageT]]],
    ) -> tuple[HashPoint[BaseT], Reducer[SequenceDispatcher[HashPoint[StageT], bool], bool]]:
        assert isinstance(element, HashPoint)
        pair: Pair[BaseT, Stack[StageT]] = await element.resolve()
        assert isinstance(pair, Pair)
        base: HashPoint[BaseT] = pair.element0
        assert isinstance(base, HashPoint)
        stages_stack: HashPoint[Stack[StageT]] = pair.element1
        assert isinstance(stages_stack, HashPoint)
        stack_bridge: Reducer[
            SequenceDispatcher[HashPoint[StageT], bool],
            bool
        ] = StackBridge(stages_stack).over_elements()
        assert isinstance(stack_bridge, Reducer)
        return base, stack_bridge


def stage_bridge(
        stage_protocol: StageProtocol[HeaderT, BaseT, StageT], base_factory: RainbowFactory[BaseT]
) -> Verification[
    tuple[
        Nullable[HashPoint[Pair[BaseT, Stack[StageT]]]],
        HashPoint[HeaderT],
        HashPoint[Pair[BaseT, Stack[StageT]]]
    ]
]:
    assert isinstance(stage_protocol, StageProtocol)
    assert isinstance(base_factory, RainbowFactory)
    vp: StageVerificationProtocol[
        HashPoint[BaseT], HashPoint[StageT], HashPoint[HeaderT]
    ] = StageBridgeVP(stage_protocol, base_factory)
    assert isinstance(vp, StageVerificationProtocol)
    m: Mapper[
        HashPoint[Pair[BaseT, Stack[StageT]]],
        tuple[HashPoint[BaseT], Reducer[SequenceDispatcher[HashPoint[StageT], bool], bool]]
    ] = StageBridgeM()
    assert isinstance(m, Mapper)
    v: Verification[
        tuple[
            Nullable[HashPoint[Pair[BaseT, Stack[StageT]]]],
            HashPoint[HeaderT],
            HashPoint[Pair[BaseT, Stack[StageT]]]
        ]
    ] = StageVerification(vp, m).loose()
    assert isinstance(v, Verification)
    return v
