from rainbowadn.core import *
from rainbowadn.flow.bridge import *
from rainbowadn.flow.primitive import *
from rainbowadn.flow.verification.core import *
from rainbowadn.nullability import *
from ._flowbank import *
from ._flowcheque import *
from ._flowstandard import *
from ._flowtransaction import *

__all__ = ('FlowBankVerification',)


class FlowBankVerification(
    Verification[tuple[Nullable[FlowBank], FlowCheque, FlowBank]],
):
    def __init__(self, initial: FlowBank):
        assert isinstance(initial, FlowBank)
        self.initial = initial

    @classmethod
    async def _verify_disjoint_union(
            cls,
            per_previous: FlowStandard[FlowCoin],
            per_cheque: FlowStandard[FlowCoin],
            per_bank: FlowStandard[FlowCoin],
    ) -> bool:
        assert isinstance(per_previous, FlowStandard)
        assert isinstance(per_cheque, FlowStandard)
        assert isinstance(per_bank, FlowStandard)

        async def verify_no_intersection():
            assert_true(await per_previous.verify_does_not_intersect(await per_cheque.reducer()))
            return True

        assert_trues(
            await gather(
                per_previous.verify_subset(UnitReducer(per_bank)),
                per_cheque.verify_subset(UnitReducer(per_bank)),
                per_bank.verify_subset(ListBridge([per_previous, per_cheque])),
                verify_no_intersection(),
            )
        )

        return True

    @classmethod
    async def _verify(
            cls,
            previous: FlowBank,
            cheque: FlowCheque,
            bank: FlowBank,
    ) -> bool:
        assert isinstance(previous, FlowBank)
        assert isinstance(cheque, FlowCheque)
        assert isinstance(bank, FlowBank)

        async def verify_unique_minted():
            previous_minted: FlowStandard[FlowCoin] = await previous.minted()
            cheque_minted: FlowStandard[FlowCoin] = await cheque.minted()
            bank_minted: FlowStandard[FlowCoin] = await bank.minted()
            assert_true(
                await cls._verify_disjoint_union(
                    previous_minted,
                    cheque_minted,
                    bank_minted,
                )
            )
            return True

        async def verify_unique_used():
            previous_used: FlowStandard[FlowCoin] = await previous.used()
            cheque_used: FlowStandard[FlowCoin] = await cheque.used()
            bank_used: FlowStandard[FlowCoin] = await bank.used()
            assert_true(
                await cls._verify_disjoint_union(
                    previous_used,
                    cheque_used,
                    bank_used,
                )
            )
            return True

        async def verify_used_were_minted():
            bank_minted: FlowStandard[FlowCoin] = await bank.minted()
            cheque_used: FlowStandard[FlowCoin] = await cheque.used()
            assert_true(
                await cheque_used.verify_subset(UnitReducer(bank_minted))
            )
            return True

        assert_trues(
            await gather(
                cheque.verify(),
                verify_unique_minted(),
                verify_unique_used(),
                verify_used_were_minted(),
            )
        )
        return True

    async def verify(
            self,
            element: tuple[Nullable[FlowBank], FlowCheque, FlowBank]
    ) -> bool:
        assert isinstance(element, tuple)
        previous: Nullable[FlowBank]
        cheque: FlowCheque
        bank: FlowBank
        previous, cheque, bank = element
        assert isinstance(previous, Nullable)
        assert isinstance(cheque, FlowCheque)
        assert isinstance(bank, FlowBank)
        previous_bank: FlowBank
        if previous.null():
            previous_bank = self.initial
        else:
            previous_bank = previous.resolve()
        assert isinstance(previous_bank, FlowBank)
        assert_true(await self._verify(previous_bank, cheque, bank))
        return True
