from rainbowadn.collection.keyvalue import *
from rainbowadn.core import *
from rainbowadn.flow.core import *
from rainbowadn.flow.verification.core import *
from ._flowstandard import *
from ._flowtransaction import *

__all__ = ('FlowCheque',)


class SumReduce(PureReduce[int]):
    async def merge(self, left: int, right: int) -> int:
        return left + right


class ValueMapper(Mapper[FlowCoin, int]):
    async def map(self, element: FlowCoin) -> int:
        assert isinstance(element, FlowCoin)
        return await element.int_value()


class FlowCheque:
    async def transactions(self) -> FlowStandard[FlowTransaction]:
        raise NotImplementedError

    async def minted(self) -> FlowStandard[FlowCoin]:
        raise NotImplementedError

    async def used(self) -> FlowStandard[FlowCoin]:
        raise NotImplementedError

    async def usedx(self) -> FlowStandard[KeyValue[FlowCoin, FlowTransaction]]:
        raise NotImplementedError

    async def mint(self) -> int:
        raise NotImplementedError

    @classmethod
    async def total_of(cls, tree: FlowStandard[FlowCoin]) -> int:
        assert isinstance(tree, FlowStandard)
        reducer: Reducer[FlowCoin] = await tree.reducer()
        assert isinstance(reducer, Reducer)
        total: int = await reducer.reduce(MapReduce(ValueMapper(), SumReduce(0)))
        assert isinstance(total, int)
        return total

    async def total_minted(self) -> int:
        return await self.total_of(await self.minted())

    async def total_used(self) -> int:
        return await self.total_of(await self.used())

    async def extra(self) -> int:
        mint: int
        total_minted: int
        total_used: int
        mint, total_minted, total_used = await gather(
            self.mint(),
            self.total_minted(),
            self.total_used(),
        )
        assert isinstance(mint, int)
        assert isinstance(total_minted, int)
        assert isinstance(total_used, int)
        return mint + total_minted - total_used

    async def _verify_extra(self) -> bool:
        assert (await self.extra()) >= 0
        return True

    async def _verify_transactions(self) -> bool:
        assert_true(
            await (await self.transactions()).verify(TransactionVerification(self).loose())
        )
        return True

    async def _verify_minted(self) -> bool:
        assert_true(
            await (await self.minted()).verify(MintedVerification(self).loose())
        )
        return True

    async def _verify_used(self) -> bool:
        assert_true(
            await (await self.used()).verify(UsedVerification(self).loose())
        )
        return True

    async def _verify_usedx(self) -> bool:
        assert_true(
            await (await self.usedx()).verify(UsedXVerification(self).loose())
        )
        return True

    async def verify(self) -> bool:
        assert_trues(
            await gather(
                self._verify_extra(),
                self._verify_transactions(),
                self._verify_minted(),
                self._verify_used(),
                self._verify_usedx(),
            )
        )
        return True


class TransactionVerification(
    Verification[FlowTransaction]
):
    def __init__(self, cheque: FlowCheque):
        assert isinstance(cheque, FlowCheque)
        self.cheque = cheque

    @classmethod
    def usedx_reducer(cls, reducer: Reducer[FlowCoin, bool], transaction: FlowTransaction):
        assert isinstance(reducer, Reducer)
        assert isinstance(transaction, FlowTransaction)

        def usedx(coin: FlowCoin) -> KeyValue[FlowCoin, FlowTransaction]:
            assert isinstance(coin, FlowCoin)
            return KeyValue(
                HashPoint.of(coin),
                HashPoint.of(transaction)
            )

        usedx_reducer: Reducer[KeyValue[FlowCoin, FlowTransaction], bool] = MapReducer(
            CallableMapper(usedx),
            reducer
        )
        assert isinstance(usedx_reducer, Reducer)
        return usedx_reducer

    async def _verify_transaction_minted(self, transaction: FlowTransaction) -> bool:
        assert isinstance(transaction, FlowTransaction)
        minted: FlowStandard[FlowCoin]
        minted_reducer: Reducer[FlowCoin, bool]
        minted, minted_reducer = await gather(
            self.cheque.minted(),
            transaction.minted_reducer(),
        )
        assert isinstance(minted, FlowStandard)
        assert isinstance(minted_reducer, Reducer)
        assert_true(
            await minted.verify_contains_all(
                minted_reducer
            )
        )
        return True

    async def _verify_transaction_used(self, transaction: FlowTransaction) -> bool:
        assert isinstance(transaction, FlowTransaction)
        minted: FlowStandard[FlowCoin]
        minted_reducer: Reducer[FlowCoin, bool]
        used, used_reducer = await gather(
            self.cheque.used(),
            transaction.used_reducer(),
        )
        assert isinstance(used, FlowStandard)
        assert isinstance(used_reducer, Reducer)
        assert_true(
            await used.verify_contains_all(
                used_reducer
            )
        )
        return True

    async def _verify_transaction_usedx(self, transaction: FlowTransaction) -> bool:
        minted: FlowStandard[KeyValue[FlowCoin, FlowTransaction]]
        minted_reducer: Reducer[FlowCoin, bool]
        usedx, used_reducer = await gather(
            self.cheque.usedx(),
            transaction.used_reducer(),
        )
        assert isinstance(usedx, FlowStandard)
        assert isinstance(used_reducer, Reducer)
        assert_true(
            await usedx.verify_contains_all(
                self.usedx_reducer(used_reducer, transaction)
            )
        )
        return True

    async def verify(self, element: FlowTransaction) -> bool:
        assert isinstance(element, FlowTransaction)
        assert_trues(
            await gather(
                self._verify_transaction_minted(element),
                self._verify_transaction_used(element),
                self._verify_transaction_usedx(element),
            )
        )
        return True

    def loose(self) -> Verification[FlowTransaction]:
        return self


class MintedVerification(
    Verification[FlowCoin]
):
    def __init__(self, cheque: FlowCheque):
        assert isinstance(cheque, FlowCheque)
        self.cheque = cheque

    async def verify(self, element: FlowCoin) -> bool:
        assert isinstance(element, FlowCoin)
        transactions: FlowStandard[FlowTransaction]
        transaction: FlowTransaction
        transactions, transaction = await gather(
            self.cheque.transactions(),
            element.transaction.resolve(),
        )
        assert isinstance(transactions, FlowStandard)
        assert isinstance(transaction, FlowTransaction)
        assert_true(
            await transactions.contains(
                transaction
            )
        )
        return True

    def loose(self) -> Verification[FlowCoin]:
        return self


class UsedVerification(
    Verification[FlowCoin]
):
    def __init__(self, cheque: FlowCheque):
        assert isinstance(cheque, FlowCheque)
        self.cheque = cheque

    async def verify(self, element: FlowCoin) -> bool:
        assert isinstance(element, FlowCoin)
        assert_true(
            await (await self.cheque.usedx()).contains(
                KeyValue(HashPoint.of(element), HashPoint.of(FlowTransaction.empty()))
            )
        )
        return True

    def loose(self) -> Verification[FlowCoin]:
        return self


class UsedXVerification(
    Verification[KeyValue[FlowCoin, FlowTransaction]]
):
    def __init__(self, cheque: FlowCheque):
        assert isinstance(cheque, FlowCheque)
        self.cheque = cheque

    async def _verify_transaction(self, transaction: FlowTransaction) -> bool:
        assert isinstance(transaction, FlowTransaction)
        assert_true(
            await (await self.cheque.transactions()).contains(
                transaction
            )
        )
        return True

    @classmethod
    async def _verify_transaction_contains(cls, transaction: FlowTransaction, coin: HashPoint[FlowCoin]) -> bool:
        assert isinstance(transaction, FlowTransaction)
        assert isinstance(coin, HashPoint)
        in_coin: HashPoint[FlowCoin]
        async for in_coin in (await transaction.data_resolved()).iter_in_coins():
            if in_coin == coin:
                return True
        raise ValueError('used coin not found in transaction')

    async def verify(self, element: KeyValue[FlowCoin, FlowTransaction]) -> bool:
        assert isinstance(element, KeyValue)
        transaction: FlowTransaction = await element.value.resolve()
        assert isinstance(transaction, FlowTransaction)
        assert_trues(
            await gather(
                self._verify_transaction(transaction),
                self._verify_transaction_contains(transaction, element.key),
            )
        )
        return True

    def loose(self) -> Verification[KeyValue[FlowCoin, FlowTransaction]]:
        return self
