from typing import Any, AsyncIterable, Iterable

import nacl.signing

from rainbowadn.atomic import *
from rainbowadn.collection.linear import *
from rainbowadn.core import *
from rainbowadn.flow.bridge import *
from rainbowadn.flow.core import *
from rainbowadn.nullability import *
from rainbowadn.v13 import *

__all__ = ('FlowCoinData', 'FlowCoin', 'FlowTransactionData', 'FlowTransaction',)


class FlowCoinData(RecursiveMentionable, StaticMentionable):
    def __init__(
            self,
            owner: HashPoint[Subject],
            value: HashPoint[Integer]
    ):
        assert isinstance(owner, HashPoint)
        assert isinstance(value, HashPoint)
        self.owner = owner
        self.value = value

    async def int_value(self) -> int:
        return (await self.value.resolve()).integer

    @classmethod
    def of(cls, owner: Subject, value: int) -> 'FlowCoinData':
        assert isinstance(owner, Subject)
        assert isinstance(value, int)
        return cls(HashPoint.of(owner), HashPoint.of(Integer(value)))

    def points(self) -> Iterable[HashPoint]:
        return [self.owner, self.value]

    def __bytes__(self):
        return bytes(self.owner) + bytes(self.value)

    @classmethod
    def from_bytes(cls, source: bytes, resolver: HashResolver) -> 'FlowCoinData':
        assert isinstance(source, bytes)
        assert isinstance(resolver, HashResolver)
        return cls(
            ResolverOrigin(Subject.factory(), source[:HashPoint.HASH_LENGTH], resolver).hash_point(),
            ResolverOrigin(Integer.factory(), source[HashPoint.HASH_LENGTH:], resolver).hash_point(),
        )

    async def str(self, tab: int) -> str:
        assert isinstance(tab, int)
        owner_str, value_str = await gather(
            hash_point_format(self.owner, tab),
            hash_point_format(self.value, tab),
        )
        assert isinstance(owner_str, str)
        assert isinstance(value_str, str)
        return f'{owner_str}' \
               f'{tabulate(tab)}{value_str}'


class FlowCoin(RecursiveMentionable, StaticMentionable):
    def __init__(
            self,
            data: HashPoint[FlowCoinData],
            transaction: HashPoint['FlowTransaction'],
            index: HashPoint[Integer]
    ):
        assert isinstance(data, HashPoint)
        assert isinstance(transaction, HashPoint)
        assert isinstance(index, HashPoint)
        self.data = data
        self.transaction = transaction
        self.index = index

    async def data_resolved(self) -> FlowCoinData:
        return await self.data.resolve()

    def points(self) -> Iterable[HashPoint]:
        return [self.data, self.transaction, self.index]

    def __bytes__(self):
        return bytes(self.data) + bytes(self.transaction) + bytes(self.index)

    @classmethod
    def from_bytes(cls, source: bytes, resolver: HashResolver) -> 'FlowCoin':
        assert isinstance(source, bytes)
        assert isinstance(resolver, HashResolver)
        return cls(
            ResolverOrigin(FlowCoinData.factory(), source[:HashPoint.HASH_LENGTH], resolver).hash_point(),
            ResolverOrigin(
                FlowTransaction.factory(), source[HashPoint.HASH_LENGTH:2 * HashPoint.HASH_LENGTH], resolver
            ).hash_point(),
            ResolverOrigin(Integer.factory(), source[2 * HashPoint.HASH_LENGTH:], resolver).hash_point(),
        )

    async def str(self, tab: int) -> str:
        assert isinstance(tab, int)
        data_str, index_str = await gather(
            hash_point_format(self.data, tab + 1),
            hash_point_format(self.index, tab + 1),
        )
        assert isinstance(data_str, str)
        assert isinstance(index_str, str)
        return f'(' \
               f'{tabulate(tab + 1)}coin' \
               f'{tabulate(tab + 1)}{data_str}' \
               f'{tabulate(tab + 1)}(origin)' \
               f'{tabulate(tab + 1)}{index_str}' \
               f'{tabulate(tab)})'

    async def int_value(self) -> int:
        return await (await self.data_resolved()).int_value()

    async def owner_resolved(self) -> Subject:
        return await (await self.data_resolved()).owner.resolve()


class FlowTransactionData(RecursiveMentionable, StaticMentionable):
    def __init__(
            self,
            in_coins: NullableReference[Stack[FlowCoin]],
            out_coins: NullableReference[Stack[FlowCoinData]],
    ):
        assert isinstance(in_coins, NullableReference)
        assert isinstance(out_coins, NullableReference)
        self.in_coins = in_coins
        self.out_coins = out_coins
        self.hash_point = HashPoint.of(self)
        assert isinstance(self.hash_point, HashPoint)

    def points(self) -> Iterable[HashPoint]:
        return [*self.in_coins.points(), *self.out_coins.points()]

    def __bytes__(self):
        return bytes(self.in_coins) + bytes(self.out_coins)

    @classmethod
    def from_bytes(cls, source: bytes, resolver: HashResolver) -> 'FlowTransactionData':
        assert isinstance(source, bytes)
        assert isinstance(resolver, HashResolver)
        return cls(
            NullableReferenceFactory(
                StackFactory(FlowCoin.factory()).loose()
            ).from_bytes(source[:HashPoint.HASH_LENGTH], resolver),
            NullableReferenceFactory(
                StackFactory(FlowCoinData.factory()).loose()
            ).from_bytes(source[HashPoint.HASH_LENGTH:], resolver),
        )

    async def _signature_verify(self, coin: FlowCoin, signature: Signature) -> bool:
        assert isinstance(coin, FlowCoin)
        assert isinstance(signature, Signature)
        assert_true(
            signature.verify(
                await coin.owner_resolved(),
                self.hash_point
            )
        )
        return True

    async def _verify_signatures(
            self,
            signatures: NullableReference[Stack[Signature]]
    ) -> bool:
        assert isinstance(signatures, NullableReference)
        assert_trues(
            await gather(
                *[
                    self._signature_verify(coin, signature)
                    for
                    coin, signature
                    in
                    zip(
                        await self.in_coins_resolved(),
                        await Stack.list(signatures),
                        strict=True
                    )
                ]
            )
        )
        return True

    def iter_in_coins(self) -> AsyncIterable[HashPoint[FlowCoin]]:
        return Stack.iter(self.in_coins)

    async def coins(self) -> list[FlowCoin]:
        return [await x.resolve() async for x in self.iter_in_coins()]

    async def in_coins_resolved(self) -> list[FlowCoin]:
        return await Stack.list(self.in_coins)

    def iter_out_coins(self) -> AsyncIterable[HashPoint[FlowCoinData]]:
        return Stack.iter(self.out_coins)

    async def out_coins_resolved(self) -> list[FlowCoinData]:
        return await Stack.list(self.out_coins)

    async def verify(
            self,
            signatures: NullableReference[Stack[Signature]]
    ) -> bool:
        assert isinstance(signatures, NullableReference)
        assert_true(await self._verify_signatures(signatures))
        return True

    async def str(self, tab: int) -> str:
        assert isinstance(tab, int)
        in_str, out_str = await gather(
            self.in_coins.str(tab),
            self.out_coins.str(tab),
        )
        assert isinstance(in_str, str)
        assert isinstance(out_str, str)
        return f'(in)' \
               f'{tabulate(tab)}{in_str}' \
               f'{tabulate(tab)}(out)' \
               f'{tabulate(tab)}{out_str}'


class FlowTransaction(RecursiveMentionable, StaticMentionable):
    def __init__(
            self,
            data: HashPoint[FlowTransactionData],
            signatures: NullableReference[Stack[Signature]]
    ):
        assert isinstance(data, HashPoint)
        assert isinstance(signatures, NullableReference)
        self.data = data
        self.signatures = signatures
        self.hash_point = HashPoint.of(self)
        assert isinstance(self.hash_point, HashPoint)

    async def data_resolved(self) -> FlowTransactionData:
        return await self.data.resolve()

    def points(self) -> Iterable[HashPoint]:
        return [self.data, *self.signatures.points()]

    def __bytes__(self):
        return bytes(self.data) + bytes(self.signatures)

    @classmethod
    def from_bytes(cls, source: bytes, resolver: HashResolver) -> 'FlowTransaction':
        assert isinstance(source, bytes)
        assert isinstance(resolver, HashResolver)
        signature_factory: RainbowFactory[Signature] = Signature.factory()
        assert isinstance(signature_factory, RainbowFactory)
        stack_factory: RainbowFactory[Stack[Signature]] = StackFactory(signature_factory).loose()
        assert isinstance(stack_factory, RainbowFactory)
        return cls(
            ResolverOrigin(FlowTransactionData.factory(), source[:HashPoint.HASH_LENGTH], resolver).hash_point(),
            NullableReferenceFactory(stack_factory).from_bytes(source[HashPoint.HASH_LENGTH:], resolver),
        )

    async def iter_coins(
            self
    ) -> AsyncIterable[FlowCoin]:
        transaction_data: FlowTransactionData = await self.data_resolved()
        assert isinstance(transaction_data, FlowTransactionData)
        index = 0
        out_coin: HashPoint[FlowCoinData]
        async for out_coin in transaction_data.iter_out_coins():
            assert isinstance(out_coin, HashPoint)
            coin: FlowCoin = FlowCoin(out_coin, self.hash_point, HashPoint.of(Integer(index)))
            assert isinstance(coin, FlowCoin)
            yield coin
            index += 1

    async def coins(
            self
    ) -> list[FlowCoin]:
        return [coin async for coin in self.iter_coins()]

    async def used_reducer(self) -> Reducer[FlowCoin, Any]:
        transaction_data: FlowTransactionData = await self.data_resolved()
        assert isinstance(transaction_data, FlowTransactionData)
        bridge: Reducer[FlowCoin, Any] = ListBridge(await transaction_data.coins())
        assert isinstance(bridge, Reducer)
        return bridge

    async def minted_reducer(self) -> Reducer[FlowCoin, Any]:
        bridge: Reducer[FlowCoin, Any] = ListBridge(await self.coins())
        assert isinstance(bridge, Reducer)
        return bridge

    async def verify(self):
        data: FlowTransactionData = await self.data_resolved()
        assert isinstance(data, FlowTransactionData)
        assert_true(await data.verify(self.signatures))
        return True

    async def str(self, tab: int) -> str:
        assert isinstance(tab, int)
        data_str, signatures_str = await gather(
            hash_point_format(self.data, tab + 1),
            self.signatures.str(tab + 1),
        )
        assert isinstance(data_str, str)
        assert isinstance(signatures_str, str)
        return f'(' \
               f'{tabulate(tab + 1)}transaction' \
               f'{tabulate(tab + 1)}{data_str}' \
               f'{tabulate(tab + 1)}{signatures_str}' \
               f'{tabulate(tab)})'

    @classmethod
    def make(
            cls,
            in_coins: list[FlowCoin],
            out_coins: list[FlowCoinData],
            keys: list[nacl.signing.SigningKey],
    ) -> 'FlowTransaction':
        assert isinstance(in_coins, list)
        assert isinstance(out_coins, list)
        assert isinstance(keys, list)
        transaction_data = FlowTransactionData(
            Stack.off(FlowCoin.factory(), reversed(in_coins)),
            Stack.off(FlowCoinData.factory(), reversed(out_coins)),
        )
        assert isinstance(transaction_data, FlowTransactionData)
        return FlowTransaction(
            HashPoint.of(transaction_data),
            Stack.off(
                Signature.factory(),
                (Signature.sign(key, HashPoint.of(transaction_data)) for key in reversed(keys))
            )
        )

    @classmethod
    def empty(cls):
        return cls.make([], [], [])
