import os
import time
import unittest
from typing import Any

from rainbowadn.atomic import *
from rainbowadn.chain import *
from rainbowadn.collection.comparison import *
from rainbowadn.collection.pair import *
from rainbowadn.collection.trees.binary import *
from rainbowadn.core import *
from rainbowadn.testing.resolvers import *
from rainbowadn.wrisbt import *


class TestTrees(unittest.IsolatedAsyncioTestCase):
    async def test_wrisbt(self):
        set_gather_linear()
        with self.subTest('setup'):
            stoptime = time.process_time()

            def measure(message: str) -> float:
                nonlocal stoptime
                now = time.process_time()
                delta = now - stoptime
                print(message, delta)
                stoptime = now
                return delta

            n = 5000
            keysize = 7
        with self.subTest('create empty'):
            btree: WrisbtRoot = WrisbtRoot.empty(WrisbtParametres(5, keysize))
            measure('init')
        with self.subTest('add keys', n=n):
            for _ in range(n):
                key = os.urandom(keysize)
                assert_false(await btree.contains(key))
                btree = await btree.add(key)
                assert_true(await btree.contains(key))
            measure('add')
        with self.subTest('save'):
            btree = await default_resolver().migrate_resolved(btree)
            measure('save')
        set_gather_asyncio()
        with self.subTest('resolve and iterate'):
            assert_eq(len(await btree.keys()), n)
            print(btree.height)
            measure('resolve and iterate')
        with self.subTest('resolve and add', n=n):
            for _ in range(n):
                key = os.urandom(keysize)
                assert_false(await btree.contains(key))
                btree = await btree.add(key)
                assert_true(await btree.contains(key))
            print(btree.height)
            measure('resolve and add')

    async def test_wrisbt_index(self):
        set_gather_linear()
        with self.subTest('create empty'):
            factory: RainbowFactory[Pair[Plain, Plain]] = PairFactory(Plain.factory(), Plain.factory()).loose()
            chain: ChainCollectionInterface[Any, Pair[Plain, Plain], WrisbtRoot] = BlockChainFactory(
                WrisbtChainProtocol(factory, 2).loose()
            ).empty().loose()
        with self.subTest('fill'):
            for _ in range(100):
                chain = await chain.add(
                    HashPoint.of(
                        Pair(
                            HashPoint.of(Plain(os.urandom(16))),
                            HashPoint.of(Plain(os.urandom(16)))
                        )
                    )
                )
        with self.subTest('check'):
            set_gather_asyncio()
            assert_true(await chain.verify())

    async def test_avl(self):
        set_gather_linear()
        tree: ActiveBinaryTree[Plain, Integer] = ActiveBinaryTree.empty(
            AVL(PlainComparator(Replace())), Plain.factory()
        )
        for i in range(26):
            tree = await tree.add(HashPoint.of(Plain(bytes([ord('A') + i]))))
        print(await tree.reference.str(0))

    async def test_avl_stress(self):
        set_gather_linear()
        protocol = AVL(PlainComparator(Replace()))
        tree: ActiveBinaryTree[Plain, Integer] = ActiveBinaryTree.empty(
            protocol, Plain.factory()
        )
        for i in range(250):
            tree = await tree.add(HashPoint.of(Plain(os.urandom(16))))
        print(await AVL.height(tree.protocolized()))
